#include "stdafx.h"
#include "spc700.h"
#include "apu.h"
#include "soundux.h"
#include "port.h"

extern int MixBuffer[/*MAX_BUFFER_SIZE*/];
extern long FilterValues[4][2];
extern int NoiseFreq[32];

#define ABS(a)	((a) < 0 ? -(a) : (a))

#define SIXTEEN_BIT_SHIFT	1

#define FIXED_POINT		0x4000
#define FIXED_POINT_REMAINDER	0x3fff
#define FIXED_POINT_SHIFT	14

void S9xSetEnvelopeRate(int channel, uint32 rate, int direction, int target)
{
	SoundData.channels[channel].envx_target = target;

	if(rate == ~0) {
		SoundData.channels[channel].direction = 0;
		rate = 0;
	} else {
		SoundData.channels[channel].direction = direction;
	}

	if(rate == 0 || so.playback_rate == 0) {
		SoundData.channels[channel].erate = 0;
	} else {
		/* ZIȂƂ */
		//SoundData.channels[channel].erate = (uint32)(((1.0 / ((double)rate / 1000.0 / MAX_ENVELOPE_HEIGHT)) * FIXED_POINT) / so.playback_rate);
		//                                  = (FIXED_POINT * MAX_ENVELOPE_HEIGHT * 1000) / (rate * so.playback_rate) ... ̎ŃWɓ邩H
		//                                  = (0x4000      * 127                 * 1000) / (rate * 16000           )
		//                                  = 0x7C060000                                 / (rate * 16000           ) ... ȂƂvłB
		SoundData.channels[channel].erate = (FIXED_POINT * MAX_ENVELOPE_HEIGHT * 1000) / (rate * so.playback_rate);
	}
}

void S9xSetSoundVolume(int channel, short volume_left, short volume_right)
{
	Channel* ch = &SoundData.channels[channel];

	if(ABS(volume_left) >= ABS(volume_right)) {
		ch->volume = volume_left;
	} else {
		ch->volume = volume_right;
	}

	SoundData.channels[channel].vol_level =
		(SoundData.channels[channel].envx *
		 SoundData.channels[channel].volume) >> VOLUME_SHIFT;
}

void S9xSetMasterVolume(short volume_left, short volume_right)
{
	if(ABS(volume_left) >= ABS(volume_right)) {
		SoundData.master_volume  = volume_left;
	} else {
		SoundData.master_volume  = volume_right;
	}
}

void S9xSetEchoVolume(short volume_left, short volume_right) {/*DUMMY*/}
void S9xSetEchoEnable(uint8 byte) {/*DUMMY*/}
void S9xSetEchoFeedback(int feedback) {/*DUMMY*/}
void S9xSetEchoDelay(int delay) {/*DUMMY*/}
void S9xSetEchoWriteEnable(uint8 byte) {/*DUMMY*/}
void S9xSetFilterCoefficient(int tap, int value) {/*DUMMY*/}
void S9xSetFrequencyModulationEnable(uint8 byte) {/*DUMMY*/}
void S9xSetSoundSample(int channel, uint16 sample_number) {/*DUMMY*/}

void S9xSetSoundKeyOff(int channel)
{
	if(SoundData.channels[channel].state != SOUND_SILENT) {
		SoundData.channels[channel].state = SOUND_RELEASE;
		S9xSetEnvelopeRate(channel, SoundData.channels[channel].release_rate, -1, 0);
		SoundData.channels[channel].mode = SOUND_MODE_NONE;
	}
}

void S9xFixSoundAfterSnapshotLoad()
{
	int i;

	/* ??? */
	for(i = 0; i < 8; i++) {
		SoundData.channels[i].needs_decode = 1;
	}
}


void S9xSetSoundADSR(int channel, int attack_rate, int decay_rate,
                   int sustain_rate, int sustain_level, int release_rate)
{
	SoundData.channels[channel].attack_rate = attack_rate;
	SoundData.channels[channel].decay_rate = decay_rate;
	SoundData.channels[channel].sustain_rate = sustain_rate;
	SoundData.channels[channel].release_rate = release_rate;
	SoundData.channels[channel].sustain_level = sustain_level;
}

void S9xSetEnvelopeHeight(int channel, int level)
{
	S9xSetEnvelopeRate(channel, ~0, 0, 0);
	SoundData.channels[channel].envx = level >> (7 - ENVELOPE_SHIFT);
	SoundData.channels[channel].vol_level =
		(SoundData.channels[channel].envx *
		 SoundData.channels[channel].volume) >> VOLUME_SHIFT;
}

int S9xGetEnvelopeHeight(int channel)
{
	if(APU.KeyedChannels & (1 << channel))
		return SoundData.channels[channel].envx;
	else
		return 0;
}

void S9xSetSoundDecayMode(int channel)
{
	Channel* ch = &SoundData.channels[channel];
	if(ch->state != SOUND_SILENT) {
		S9xSetEnvelopeRate(channel, ch->decay_rate, -1, 0);
		ch->state = SOUND_DECAY;
	}
}

int S9xSetSoundMode(int channel, int mode)
{
	switch(mode) {
	case SOUND_MODE_ADSR:
		if(SoundData.channels[channel].mode == SOUND_MODE_NONE)
			SoundData.channels[channel].mode = SOUND_MODE_ADSR;
		break;
	case SOUND_MODE_GAIN:
		switch(SoundData.channels[channel].mode) {
		case SOUND_MODE_NONE:
		case SOUND_MODE_INCREASE:
		case SOUND_MODE_DECREASE:
		case SOUND_MODE_ADSR:
			SoundData.channels[channel].mode = SOUND_MODE_GAIN;
			break;
		}
		break;
	case SOUND_MODE_INCREASE:
		if(SoundData.channels[channel].mode == SOUND_MODE_NONE ||
		   SoundData.channels[channel].mode == SOUND_MODE_DECREASE)
			SoundData.channels[channel].mode = SOUND_MODE_INCREASE;
		break;
	case SOUND_MODE_DECREASE:
		SoundData.channels[channel].mode = SOUND_MODE_DECREASE;
		break;
	case SOUND_MODE_NONE:
		SoundData.channels[channel].mode = SOUND_MODE_NONE;
		break;
	}
	return SoundData.channels[channel].mode;
}

void S9xSetSoundFrequency(int channel, int hertz)
{
	SoundData.channels[channel].frequency = hertz;
}

void S9xSetSoundType(int channel, int type_of_sound)
{
	SoundData.channels[channel].type = type_of_sound;
}

void S9xSetSoundMute(bool8 mute)
{
	so.mute_sound = mute;
}

void DecodeBlock(Channel* ch)
{
	long out;
	unsigned char filter;
	unsigned char shift;
	char sample1, sample2;
	unsigned char i;
	char* compressed;
	short *raw;
	long prev0;
	long prev1;
	long* f;

	if(ch->block_pointer > 0x10000 - 9) {
		ch->last_block = 1;
		ch->loop = 0;
		return;
	}

	compressed = (char*)&IAPU.RAM[ch->block_pointer];
	filter = *compressed;
	if(filter & 1) {
		ch->last_block = 1;
		if(filter & 2) ch->loop = 1;
	}
	compressed++;
	raw = ch->block;
	shift  =  filter >> 4;
	filter = (filter >> 2) & 3;
	prev0 = ch->previous[0];
	prev1 = ch->previous[1];
	f = FilterValues[filter];

	/* 16Tv2TvÂWJB */
	for(i = 0; i < 16 / 2; i++) {
		sample1 = *compressed++;
		sample2 = sample1 << 4;
		//Sample 2 = Bottom Nibble, Sign Extended.
		sample2 >>= 4;
		//Sample 1 = Top Nibble, shifted down and Sign Extended.
		sample1 >>= 4;

		out = (sample1 << shift) + (((prev0 * f[0]) + (prev1 * f[1])) >> 8);
		if(out < -32768) {
			out = -32768;
		} else if(out > 32767) {
			out = 32767;
		}
		prev1 = prev0;
		prev0 = out;
		*raw++ = (short)out;

		out = (sample2 << shift) + (((prev0 * f[0]) + (prev1 * f[1])) >> 8);
		if(out < -32768) {
			out = -32768;
		} else if(out > 32767) {
			out = 32767;
		}
		prev1 = prev0;
		prev0 = out;
		*raw++ = (short) out;
	}
	ch->previous[0] = (short)prev0;
	ch->previous[1] = (short)prev1;

	if((ch->block_pointer += 9) >= 0x10000 - 9) {
		ch->last_block = 1;
		ch->loop = 0;
		ch->block_pointer -= 0x10000 - 9;
	}
}

int envelope_shift;

void MixMono(int sample_count)
{
	Channel* ch;
	int I, J, K, V, freq;
	uint32 step;
	uint8* dir;

	envelope_shift = ENVELOPE_SHIFT - g_VM;

	/* e`lɂ... */
	for(J = 0; J < NUM_CHANNELS; J++) {
//	for(J = 0; J < 4; J++) {
		ch = &SoundData.channels[J];
		if(ch->state == SOUND_SILENT) goto SKIP; /* ĂȂΔ΂ */

		if(ch->type == SOUND_NOISE) {
			freq = NoiseFreq[APU.DSP[APU_FLG] & 0x1f];
		} else {
			freq = ch->frequency;
		}

		if(freq) {
			/* ZIȂƂ */
			//K = (uint32)(((double)freq + 1) * FIXED_POINT / so.playback_rate);
			//  = (freq + 1) * FIXED_POINT / so.playback_rate ... ̎ŃWɓ邩H
			//  StartSample()琄freq̍ől(FREQUENCY_MASK*8)Ȃ̂ŁA
			//  = (0x3fff * 8 + 1) * 0x4000 / so.playback_rate
			//  = 0x7ffe4000 / so.playback_rate ... ȂƂvłB
			K = (freq + 1) * FIXED_POINT / so.playback_rate;
		} else {
			K = 0;
		}

		/* TvOf[^ADPCM`łB
		 * 16Tv9oCgƂ16Tv32oCgɓWJ܂B(1+4bit*16) -> (16bit*16)
		 */
		if(ch->needs_decode) {
			DecodeBlock(ch);
			ch->needs_decode = 0;
			ch->sample_pointer = 1;
			ch->sample = ch->block[0];
		}

		if(ch->type == SOUND_NOISE) {
			V = (((so.noise_gen << 17) >> 17) * ch->vol_level) >> envelope_shift;
		} else {
			V = (ch->sample * ch->vol_level) >> envelope_shift;
		}

		for(I = 0; I < sample_count; I++) {
			ch->env_error += ch->erate;
			if(ch->env_error >= FIXED_POINT) {
				step = (ch->env_error >> FIXED_POINT_SHIFT) - 1;
				if(step) {
					if(ch->direction < 0) {
						if(ch->envx - (int) step <= ch->envx_target)
							ch->envx = ch->envx_target;
						else
							ch->envx -= step;
					} else {
						if(ch->envx + (int) step >= ch->envx_target)
							ch->envx = ch->envx_target;
						else
							ch->envx += step;
					}
					ch->env_error &= FIXED_POINT_REMAINDER;
				} else {
					ch->env_error -= FIXED_POINT;
				}
				if(ch->envx == ch->envx_target || ch->envx < 0)
				{
					if(ch->envx <= 0) {
						ch->state = SOUND_SILENT;
						S9xAPUSetEndOfSample (J);
						APU.KeyedChannels &= ~(1 << J);
						goto SKIP;
					} else {
						switch(++ch->state) {
						case SOUND_DECAY:
							if(ch->sustain_level != 7) {
								S9xSetEnvelopeRate(J, ch->decay_rate, -1, (MAX_ENVELOPE_HEIGHT * (ch->sustain_level + 1)) >> 3);
								break;
							}
							ch->state++;
							// Fall ...
						case SOUND_SUSTAIN:
							S9xSetEnvelopeRate(J, ch->sustain_rate, -1, 0);
							break;
						case SOUND_RELEASE:
							S9xSetEnvelopeRate(J, ch->release_rate, -1, 0);
							break;
						default:
							ch->state = SOUND_SILENT;
							S9xAPUSetEndOfSample(J);
							APU.KeyedChannels &= ~(1 << J);
							goto SKIP;
						}
					}
				}
				else {
					ch->envx += ch->direction;
				}
				ch->vol_level = (ch->envx * ch->volume) >> VOLUME_SHIFT;
			}

			ch->count += K;
			if(ch->count >= FIXED_POINT) {
				uint32 step = ch->count >> FIXED_POINT_SHIFT;
				ch->count &= FIXED_POINT_REMAINDER;
				ch->sample_pointer += step - 1;
				if(ch->sample_pointer >= 16) {
					ch->sample_pointer -= 16;
					if(ch->last_block) {
						S9xAPUSetEndX(J);
						if(!ch->loop) {
							ch->state = SOUND_SILENT;
							APU.KeyedChannels &= ~(1 << J);
							S9xAPUSetEndOfSample (J);
							goto SKIP;
						}
						ch->last_block = 0;
						dir = IAPU.RAM + (((APU.DSP[APU_DIR] << 8) + (ch->sample_number << 2)) & 0xffff);
						ch->block_pointer = *(dir + 2) + (*(dir + 3) << 8);
					}
					DecodeBlock (ch);
				}
				ch->sample = ch->block[ch->sample_pointer++];
				if(ch->type == SOUND_NOISE) {
					if((so.noise_gen <<= 1) & 0x80000000L) so.noise_gen ^= 0x0040001L;
					V = (((so.noise_gen << 17) >> 17) * ch->vol_level) >> envelope_shift;
				} else {
					V = (ch->sample * ch->vol_level) >> envelope_shift;
				}
			}
			MixBuffer [I] += V;
		}
SKIP: ;
	}
}

void S9xMixSamples(short* buffer, int sample_count)
{
	int I, J;

	if(so.mute_sound) {
		for(J = 0; J < sample_count; J++) buffer[J] = 0;
	} else {
		/* B */
		memset(MixBuffer, 0, sample_count * sizeof *MixBuffer);
		/* mŊe`l~LVOB */
		MixMono(sample_count);
		/* 16rbgEmEGR[ȂŌ㏈B */
		for(J = 0; J < sample_count; J++) {
			I = MixBuffer[J] >> SIXTEEN_BIT_SHIFT;
			I = (I * SoundData.master_volume) >> VOLUME_SHIFT;
			if(I > 32767) {
				I = 32767;
			} else if(I < -32768) {
				I = -32768;
			}
			buffer[J] = I;
		}
	}
}

void S9xResetSound()
{
	int i;
	for(i = 0; i < 8; i++){
		SoundData.channels[i].state = SOUND_SILENT;
		SoundData.channels[i].type = SOUND_SAMPLE;
		SoundData.channels[i].volume = 0;
		SoundData.channels[i].frequency = 0;
		SoundData.channels[i].count = 0;
		SoundData.channels[i].loop = 0;
		SoundData.channels[i].envx_target = 0;
		SoundData.channels[i].env_error = 0;
		SoundData.channels[i].erate = 0;
		SoundData.channels[i].envx = 0;
		SoundData.channels[i].vol_level = 0;
		SoundData.channels[i].direction = 0;
		SoundData.channels[i].attack_rate = 0;
		SoundData.channels[i].decay_rate = 0;
		SoundData.channels[i].sustain_rate = 0;
		SoundData.channels[i].release_rate = 0;
		SoundData.channels[i].sustain_level = 0;
		SoundData.channels[i].mode = SOUND_MODE_NONE;
	}
	so.mute_sound = 1;
	so.noise_gen = 1;
}

void S9xPlaySample(int channel, int type, short volume_left, short volume_right, int frequency, uint16 sample_number)
{
	Channel* ch = &SoundData.channels[channel];
	uint8* dir;

	/*  */
	ch->state = SOUND_SILENT;			/* ܂܂ */
	ch->type = type;				/* TvO/mCYWFl[^I */
	if(ABS(volume_left) >= ABS(volume_right)) {	/* {[ݒ */
		ch->volume = volume_left;
	} else {
		ch->volume = volume_right;
	}
	ch->frequency = frequency;			/* gݒ */
	ch->sample_number = sample_number;		/* TvỎFݒ */
	ch->loop = 0;				/* [v͉Ff[^Ɋ܂܂̂ŁADecodeBlock()Őݒ肵܂ */
	ch->needs_decode = 1;				/* ܂Ff[^WJĂȂ̂ŁAWJKv */
	ch->last_block = 0;				/* ŏIubNł͂Ȃ */
	ch->previous[0] = ch->previous[1] = 0;		/* WJpf[^NA */
	dir = IAPU.RAM + (((APU.DSP[APU_DIR] << 8) + (sample_number << 2)) & 0xffff);	/* Ff[^ւ̃|C^̃AhX */
	ch->block_pointer = *dir + (*(dir + 1) << 8);	/* Ff[^̐擪AhX */
	ch->sample_pointer = 0;				/* WJς݉Ff[^̑ItZbg */
	ch->env_error = 0;				/* Gx[vDDǍ덷 */

	/* Jn */
	if(ch->attack_rate == 0) {
		if(ch->decay_rate == 0 || ch->sustain_level == 7) {
			if(ch->mode != SOUND_MODE_GAIN) {
				ch->envx = (MAX_ENVELOPE_HEIGHT * (ch->sustain_level + 1)) >> 3;
			}
			ch->vol_level = (ch->envx * ch->volume) >> VOLUME_SHIFT;
			S9xSetEnvelopeRate(channel, ch->sustain_rate, -1, 0);
			ch->state = SOUND_SUSTAIN;
		} else {
			ch->envx = MAX_ENVELOPE_HEIGHT;
			ch->vol_level = (ch->envx * ch->volume) >> VOLUME_SHIFT;
			S9xSetEnvelopeRate(channel, ch->decay_rate, -1, (MAX_ENVELOPE_HEIGHT * (ch->sustain_level + 1)) >> 3);
			ch->state = SOUND_DECAY;
		}
	} else {
		ch->envx = 0;
		ch->vol_level = 0;
		S9xSetEnvelopeRate(channel, ch->attack_rate, 1, MAX_ENVELOPE_HEIGHT);
		ch->state = SOUND_ATTACK;
	}
}

/****************************************************************/
extern uint8 APUROM[64];

uint8 S9xAPUGetByteZ(uint8 Address)
{
	uint8 t;

	/* DSP̓ǂݏo */
	if(Address == 0xf3) return S9xGetAPUDSP();
	/* ^C}荞݃tO擾ANA */
	if(Address >= 0xfd && IAPU.DirectPage == IAPU.RAM) {
		t = IAPU.RAM[Address];
		IAPU.RAM[Address] = 0;
		return t;
	}
	/* ʏRAM */
	return IAPU.DirectPage[Address];
}

void S9xAPUSetByteZ(uint8 byte, uint8 Address)
{
	if(Address >= 0xf0 && IAPU.DirectPage == IAPU.RAM) {
		/* DSPւ̏ */
		if(Address == 0xf3) {
			S9xSetAPUDSP(byte);
		/* o̓|[g */
		} else if(Address >= 0xf4 && Address <= 0xf7) {
			APU.OutPorts[Address - 0xf4] = byte;
		/* APU */
		} else if(Address == 0xf1) {
			S9xSetAPUControl(byte);
		/* ^CAbvlݒ */
		} else if(Address < 0xfd) {
			IAPU.RAM[Address] = byte; /* Ƃ肠i[H */
			if(Address >= 0xfa) { /* 0xfa`0xfcȂ^CAbvlݒ */
				if(byte == 0) {
					APU.TimerTarget[Address - 0xfa] = 0x100;
				} else {
					APU.TimerTarget[Address - 0xfa] = byte;
				}
			}
		}
	} else {
		/* ʏRAM */
		IAPU.DirectPage[Address] = byte;
	}
}

uint8 S9xAPUGetByte(uint16 Address)
{
	uint8 t;

	/* DSP̓ǂݏo */
	if(Address == 0xf3) return S9xGetAPUDSP();
	/* ^C}荞݃tO擾ANA */
	if(Address <= 0xff && Address >= 0xfd) {
		t = IAPU.RAM[Address];
		IAPU.RAM[Address] = 0;
		return t;
	}
	/* ROM܂͊gRAM */
	if(Address >= 0xffc0) {
		if(APU.ShowROM) {
			return APUROM[Address - 0xffc0];
		} else {
			return APU.ExtraRAM[Address - 0xffc0];
		}
	}
	/* ʏRAM */
	return IAPU.RAM[Address];
}

void S9xAPUSetByte(uint8 byte, uint16 Address)
{
	if(Address <= 0xff && Address >= 0xf0) { /* 𑁂߂悤u<=0xffvɌ */
		/* DSPւ̏ */
		if(Address == 0xf3) {
			S9xSetAPUDSP(byte);
		/* o̓|[g */
		} else if(Address >= 0xf4 && Address <= 0xf7) {
			APU.OutPorts[Address - 0xf4] = byte;
		/* APU */
		} else if(Address == 0xf1) {
			S9xSetAPUControl(byte);
		/* ^CAbvlݒ */
		} else if(Address < 0xfd) {
			IAPU.RAM[Address] = byte; /* Ƃ肠i[H */
			if(Address >= 0xfa) { /* 0xfa`0xfcȂ^CAbvlݒ */
				if(byte == 0) {
					APU.TimerTarget[Address - 0xfa] = 0x100;
				} else {
					APU.TimerTarget[Address - 0xfa] = byte;
				}
			}
		}
	} else {
		/* ʏRAM */
		if(Address < 0xffc0) {
			IAPU.RAM[Address] = byte;
		/* gRAM */
		} else {
			APU.ExtraRAM[Address - 0xffc0] = byte;
		}
	}
}


#define APU_START(OP) void Apu##OP() {
#define APU_END }
#define get16(p)	(*((uint8*)(p) + 0) << 0 | \
			 *((uint8*)(p) + 1) << 8)
#define Absolute()\
	IAPU.Address = get16(IAPU.PC + 1);
#define APUSetZN8(b)	IAPU._Zero = (b);

APU_START(0xEC)
// MOV Y,abs
Absolute();
APURegisters.YA.B.Y = S9xAPUGetByte(IAPU.Address);
APUSetZN8(APURegisters.YA.B.Y);
IAPU.PC += 3;
APU_END
