#ifndef __SOUND_H__
#define __SOUND_H__

enum {SOUND_SAMPLE,SOUND_NOISE};
enum {SOUND_SILENT,SOUND_ATTACK,SOUND_DECAY,SOUND_SUSTAIN,SOUND_RELEASE};
enum {SOUND_MODE_NONE,SOUND_MODE_ADSR,SOUND_MODE_GAIN,SOUND_MODE_INCREASE,SOUND_MODE_DECREASE};

#define NUM_CHANNELS		  8		/* SPCdl */
#define MAX_ENVELOPE_HEIGHT	127
#define ENVELOPE_SHIFT		  7
#define MAX_VOLUME		127
#define VOLUME_SHIFT		  7

#define MAX_BUFFER_SIZE		(1024 * 16)	/*   */

typedef struct _SoundStatus {
	int playback_rate;	/* o̓foCX̍Đ[g */
	int noise_gen;		/* mCYWFl[^̗V[h */
	bool8 mute_sound;	/* ~[gtO */
} SoundStatus;
extern SoundStatus so;

typedef struct _Channel {
	int state;
	int type;
	short volume;
	int frequency;
	int count;
	bool8 loop;
	int envx;
	short vol_level;
	short envx_target;
	uint32 env_error;
	uint32 erate;
	int direction;
	uint32 attack_rate;
	uint32 decay_rate;
	uint32 sustain_rate;
	uint32 release_rate;
	uint32 sustain_level;
	short sample;
	short block[16];
	short previous[2];
	uint16 sample_number;
	bool8 last_block;
	bool8 needs_decode;
	uint32 block_pointer;
	uint32 sample_pointer;
	int mode;
} Channel;

typedef struct _SSoundData {
	short master_volume;
	Channel channels[NUM_CHANNELS];
} SSoundData;
extern SSoundData SoundData;

void S9xSetSoundVolume(int channel, short volume_left, short volume_right);
void S9xSetSoundFrequency(int channel, int hertz);
void S9xSetSoundType(int channel, int type_of_sound);
void S9xSetMasterVolume(short master_volume_left, short master_volume_right);
void S9xSetEchoVolume(short echo_volume_left, short echo_volume_right);
void S9xSetSoundMute(bool8 mute);
void S9xSetEnvelopeHeight(int channel, int height);
void S9xSetSoundADSR(int channel, int attack, int decay, int sustain, int sustain_level, int release);
void S9xSetSoundKeyOff(int channel);
void S9xSetSoundDecayMode(int channel);
void S9xSetSoundAttachMode(int channel);
int S9xSetSoundMode(int channel, int mode);
void S9xSetSoundSample(int channel, uint16 sample_number);
void S9xSetEchoFeedback(int echo_feedback);
void S9xSetEchoEnable(uint8 byte);
void S9xSetEchoDelay(int byte);
void S9xSetEchoWriteEnable(uint8 byte);
void S9xSetFilterCoefficient(int tap, int value);
void S9xSetFrequencyModulationEnable(uint8 byte);
void S9xSetEnvelopeRate(int channel, uint32 rate, int direction, int target);

int S9xGetEnvelopeHeight(int channel);
void S9xResetSound();
void S9xFixSoundAfterSnapshotLoad();

void S9xPlaySample(int channel, int type, short volume_left, short volume_right, int frequency, uint16 sample_number);
void S9xMixSamples(short* buffer, int sample_count);

#endif /*__SOUND_H__*/
